(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2014-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: pixmap.ml,v 1.11 2014/02/27 16:45:22 fclement Exp $ *)

(* Pixmap: extra graphics functions primitives: handling pixels and
   pixmaps with HSB and RGB models of colors. *)

type saturation_threshold = int and brightness_threshold = int;;

let draw_pixmap pixmap x y =
 let img = Graphics.make_image pixmap in
 Graphics.draw_image img x y;;

(* Sets each pixel [p] of an image with the value [f p]. *)
let set_pixmap_pixels f px =
 let l = Array.length px in
 if l <> 0 then
 let c = Array.length px.(0) in
 for i = 0 to l - 1 do
  let line = px.(i) in
  for j = 0 to c - 1 do
   line.(j) <- f line.(j)
  done
 done;;

(* maps [f] on [px] and returns the resulting pixmap. *)
let map f px =
 let l = Array.length px in
 if l = 0 then [||] else
 let c = Array.length px.(0) in
 let res = Array.make_matrix l c 0 in
 for i = 0 to l - 1 do
  let line = px.(i) in
  let res_line = res.(i) in
  for j = 0 to c - 1 do
   res_line.(j) <- f line.(j)
  done
 done;
 res;;

let pixel_is_set ts tb c =
 let _h, s, b = Colors.hsb_of_color c in
 s >= ts || b >= tb;;

let bitmap_of_pixmap ts tb px =
 let f c = if pixel_is_set ts tb c then Graphics.white else Graphics.black in
 map f px;;

let bluemap_of_red_channel px = map Colors.red_of_color px;;
let bluemap_of_green_channel px = map Colors.green_of_color px;;
let bluemap_of_blue_channel px = map Colors.blue_of_color px;;

let bluemap_of_hue_channel px = map Colors.hue_of_color px;;
let bluemap_of_saturation_channel px = map Colors.saturation_of_color px;;
let bluemap_of_brightness_channel px = map Colors.brightness_of_color px;;

let pixmap_of_rgb_bluemaps rgp ggp bgp =
  let v = Array.map Array.copy bgp in
  for i = 0 to Array.length v - 1 do
    let liv = v.(i) in
    let lir = rgp.(i) in
    let lig = ggp.(i) in
    let lib = bgp.(i) in
    for j = 0 to Array.length liv - 1 do
      liv.(j) <- Colors.color_of_rgb lir.(j) lig.(j) lib.(j)
    done
  done;
  v;;

let pixmap_of_hsb_bluemaps hgp sgp bgp =
  let v = Array.map Array.copy bgp in
  for i = 0 to Array.length v - 1 do
    let liv = v.(i) in
    let lih = hgp.(i) in
    let lis = sgp.(i) in
    let lib = bgp.(i) in
    for j = 0 to Array.length liv - 1 do
      liv.(j) <- Colors.color_of_hsb lih.(j) lis.(j) lib.(j)
    done
  done;
  v;;

let gray_pixmap_of_bluemap bmp = map (fun c -> Colors.color_of_rgb c c c) bmp;;
let red_pixmap_of_bluemap bmp = map (fun c -> Colors.color_of_rgb c 0 0) bmp;;
let green_pixmap_of_bluemap bmp = map (fun c -> Colors.color_of_rgb 0 c 0) bmp;;
let blue_pixmap_of_bluemap bmp = bmp;;

let red_pixmap_of_red_channel px =
  map (fun c -> Colors.color_of_rgb (Colors.red_of_color c) 0 0) px
;;
let green_pixmap_of_green_channel px =
  map (fun c -> Colors.color_of_rgb 0 (Colors.green_of_color c) 0) px
;;
let blue_pixmap_of_blue_channel px =
  map (fun c -> Colors.color_of_rgb 0 0 (Colors.blue_of_color c)) px
;;

let rgb_pixmap_of_rgb_channels pxr pxg pxb =
  let v = Array.map Array.copy pxb in
  for i = 0 to Array.length v - 1 do
    let liv = v.(i) in
    let lir = pxr.(i) in
    let lig = pxg.(i) in
    let lib = pxb.(i) in
    for j = 0 to Array.length liv - 1 do
      liv.(j) <-
        Colors.color_of_rgb
          (Colors.red_of_color lir.(j))
          (Colors.green_of_color lig.(j))
          (Colors.blue_of_color lib.(j))
    done
  done;
  v
;;

(* Set to transparent color all the pixels of an image which verifies
   predicate [p]. *)

let set_pixels_transp p =
 let f c = if p c then Graphics.transp else c in
 set_pixmap_pixels f;;

(* Set to transparent the pixels of an image which have the background
  color as color. *)

let set_background_transp px =
 let f c = if c = Graphics.background then Graphics.transp else c in
 set_pixmap_pixels f px;;

(* Conversions from pixmaps to generic images. *)

open Generic_image;;

let pixmap_of_generic_image = function
 {bitmap = bm; width = w; height = h } ->
   match bm with
   | Bpp8 {data = b; colormap = colormap; transparent = _} ->
      let cmap =
        Array.map (fun {r = r; g = g; b = b} -> Graphics.rgb r g b) colormap in
      Array.init h (fun i ->
      Array.init w (fun j -> cmap.(int_of_char b.[i * w + j])))
   | Bpp24 b ->
      Array.init h (fun i ->
      Array.init w (fun j ->
       let c = 3 * (i * w + j) in
       let r = int_of_char b.[c] in
       let g = int_of_char b.[c + 1] in
       let b = int_of_char b.[c + 2] in
       Graphics.rgb r g b));;

let rgb_chars_of_color c =
 let r = Colors.red_of_color c
 and g = Colors.green_of_color c
 and b = Colors.blue_of_color c in
 char_of_int r, char_of_int g, char_of_int b;;

(* Trivial conversion: should analyse the pixmap to create a grimage
   with a [Bpp8] bitmap when it is possible. *)
let generic_image_of_pixmap img =
 let h = Array.length img in
 let w = if h = 0 then 0 else Array.length img.(0) in
 let bitmap =
  if h = 0 then "" else begin
   let s = String.create (3 * h * w) in
   let k = ref 0 in
   for i = 0 to h - 1 do
    let line = img.(i) in
    for j = 0 to w - 1 do
     let r, g, b = rgb_chars_of_color line.(j) in
     s.[!k] <- r; incr k;
     s.[!k] <- g; incr k;
     s.[!k] <- b; incr k;
    done;
   done;
   s end in
 { bitmap = Bpp24 bitmap;
   width = w; height = h };;

let image_of_pixmap = Graphics.make_image;;
let pixmap_of_image = Graphics.dump_image;;

let generic_image_of_image img =
  generic_image_of_pixmap (pixmap_of_image img);;
let image_of_generic_image gimg =
  image_of_pixmap (pixmap_of_generic_image gimg);;
