/***********************************************************************/
/*                                                                     */
/*                               Pio (C)                               */
/*                                                                     */
/*                    Polyglot Input/Output Library                    */
/*                                                                     */
/*               EPI Pomdapi, INRIA Paris - Rocquencourt               */
/*                                                                     */
/*  Copyright 2011-2014 INRIA.                                         */
/*  Institut National de Recherche en Informatique et en Automatique.  */
/*  All rights reserved. This file is distributed only by permission.  */
/*                                                                     */
/*  Bruno Verlyck <Bruno.Verlyck@inria.fr>                             */
/*                                                                     */
/***********************************************************************/

#include "common.h"
#include "value_type.h"
#include "value_type_def.h"

#define BOOL_TAG       "%B"
#define STRING_TAG     "%S"
#define INT_TAG	       "%i"
#define NATIVE_INT_TAG "%ni"
#define INT32_TAG      "%li"
#define INT64_TAG      "%Li"
#define FLOAT_BIN_TAG  "%bf"
#define FLOAT_ASCII_TAG "%f"

#define COUPLEIZE(tag) ("(" tag ", " tag ")")
#define TRIPLEIZE(tag) ("(" tag ", " tag ", " tag ")")

#define funname "::scalar_type_to_ob: "
static void scalar_type_to_ob (pio_buffer *ob, vtd_scalar_type vst) {
  char *tag = NULL;
  switch (vst) {
  default:
    error (funname, "unimpl. scalar type ?");
  case VTD_Bool_type:		tag = BOOL_TAG; break;
  case VTD_String_type:		tag = STRING_TAG; break;
  case VTD_Int_type:		tag = INT_TAG; break;
  case VTD_Native_Int_type:	tag = NATIVE_INT_TAG; break;
  case VTD_Int32_type:		tag = INT32_TAG; break;
  case VTD_Int64_type:		tag = INT64_TAG; break;
  case VTD_Float_bin_type:	tag = FLOAT_BIN_TAG; break;
  case VTD_Float_type:		tag = binary_mode ? FLOAT_BIN_TAG : FLOAT_ASCII_TAG; break;
  case VTD_Couple_Bool_type:	tag = COUPLEIZE (BOOL_TAG); break;
  case VTD_Couple_Int_type:	tag = COUPLEIZE (INT_TAG); break;
  case VTD_Couple_Native_Int_type:
				tag = COUPLEIZE (NATIVE_INT_TAG); break;
  case VTD_Couple_Int32_type:	tag = COUPLEIZE (INT32_TAG); break;
  case VTD_Couple_Int64_type:   tag = COUPLEIZE (INT64_TAG); break;
  case VTD_Couple_Float_type:	if (ascii_mode) {
				  tag = COUPLEIZE (FLOAT_ASCII_TAG);
				  break; /* in case the next else disappears */
				} /* More or less falls through */
				else	/* Only Stroustrup can suggest you such crap. */
  case VTD_Couple_Float_bin_type:
				  tag = COUPLEIZE (FLOAT_BIN_TAG);
				break;
  case VTD_Triple_Bool_type:	tag = TRIPLEIZE (BOOL_TAG); break;
  case VTD_Triple_Int_type:	tag = TRIPLEIZE (INT_TAG); break;
  case VTD_Triple_Native_Int_type:
				tag = TRIPLEIZE (NATIVE_INT_TAG); break;
  case VTD_Triple_Int32_type:	tag = TRIPLEIZE (INT32_TAG); break;
  case VTD_Triple_Int64_type:   tag = TRIPLEIZE (INT64_TAG); break;
  case VTD_Triple_Float_type:	if (ascii_mode) {
				  tag = TRIPLEIZE (FLOAT_ASCII_TAG);
				  break;
				} /* Falls through */
  case VTD_Triple_Float_bin_type: 
				tag = TRIPLEIZE (FLOAT_BIN_TAG); break;
  }
  buf_add_chars (ob, tag);
}
#undef funname

#define funname "::vt_to_ob: "
void vt_to_ob (pio_buffer *ob, vtd_value_type vt)
{
  switch (vtd_check_value_type (vt)) {
  case VTD_SCALAR_TYPE:
    scalar_type_to_ob (ob, vtd_scalar_type_of_value_type (vt));
    break;
  case VTD_VECTOR_TYPE:
    buf_add_chars (ob, "[1");
    scalar_type_to_ob (ob, vtd_scalar_type_of_value_type (vt));
    buf_add_chars (ob, "1]");
    break;
  case VTD_MATRIX_TYPE:
    buf_add_chars (ob, "[2");
    scalar_type_to_ob (ob, vtd_scalar_type_of_value_type (vt));
    buf_add_chars (ob, "2]");
    break;
  default:
    assert (0);				/* => vtd_check_value_type bug */
  }
}
#undef funname

#define funname "::scalar_type_of_ib: "
static vtd_scalar_type scalar_type_of_ib (pio_input_buffer *ib) {
  char pfx;
  ibuf_scanf (ib, 1, "%c", &pfx);
  switch (pfx) {
  case '%': {
    char type;
    ibuf_scanf (ib, 1, "%c", &type);
    switch (type) {
    case 'B': return VTD_Bool_type;
    case 'S': return VTD_String_type;
    case 'i': return VTD_Int_type;
    case 'f': return VTD_Float_type;
    case 'b': {
      char ftyp;
      ibuf_scanf (ib, 1, "%c", &ftyp); /* We differ here from the ml implementation */
      if (ftyp == 'f')
	return VTD_Float_bin_type;
      failwithf (__FILE__ funname "unknown binary format %%b%c", ftyp); /* throw */
    }
    default:
      failwithf ( __FILE__ funname "unknown format %%%c", type); /* throw */
    } /* switch (type) */
    break;
  } /* case '%' */
  case '(': {
    vtd_scalar_type types[3];
    types[0] = scalar_type_of_ib (ib);
    size_t n = 1;			/* size of tuple */
    while (true) {
      char sep;
      ibuf_scanf (ib, 1, "%c", &sep);
      switch (sep) {
      case ')':
	switch (n) {
	case 1:
	  return types[0];
	case 2:
	  return make_couple_type (types[0], types[1]);
	case 3:
	  return make_triple_type (types[0], types[1], types[2]);
	default:
	  error (__FILE__ funname, "tuple are restricted to couple or triple");
	} /* switch (n) */
	break;
      case ',':
	ibuf_scanf0 (ib, " ");		/* value_type.ml accepts only one space here */
	vtd_scalar_type t = scalar_type_of_ib (ib);
	if (++n <= 3)
	  types[n - 1] = t;		/* only keep the first 3 */
	break;
      default:
	failwithf (__FILE__ funname "',' or ')' expected; '%c' found", sep); /* throw */
      }	/* switch (sep) */
    }	/* while (true) */
  }	/* case '(' */
    break;
  default:
    failwithf (__FILE__ funname "'%%' or '(' expected; '%c' found", pfx); /* throw */
  } /* switch (pfx) */
}
#undef funname

#define funname "::vt_of_ib: "
vtd_value_type vt_of_ib (pio_input_buffer *ib) {
  char pfx;
  ibuf_scanf (ib, 1, "%c", &pfx);	/* %0c fait %1c contrairement  OCaml */
  if (pfx == '[') {
      int dim;
      vtd_value_type t;
      ibuf_scanf (ib, 1, "%d",  &dim);
      switch (dim) {
      case 1:
	t = vtd_vector_type (scalar_type_of_ib (ib));
	ibuf_scanf0 (ib, "1]");
	return t;
      case 2:
	t = vtd_matrix_type (scalar_type_of_ib (ib));
	ibuf_scanf0 (ib, "2]");
	return t;
      default:
	failwithf (__FILE__ funname	/* throw */
		   "multi-dimensional array of dimension %i not yet implemented",
		   dim);
      }
  } else {
    --ib->IB_buf.buf_position;		/* fake a %0c */
    return vtd_scalar_type (scalar_type_of_ib (ib));
  }
}
#undef funname
