(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: worker.ml,v 1.12 2014/06/25 14:55:52 fclement Exp $ *)

(* Generic part of workers for the inversion platform. *)

module Nat = Unatural;;

module Cmd_opt = Ucommand_option;;

open Misc_sprint;;
open Wire_types;;

let w_qual = Upervasives.qualify "Worker";;

let w_invalid_arg = Upervasives.invalid_arg_qual w_qual;;
let w_failwith = Upervasives.failwith_qual w_qual;;

(* Tasks. *)

type init_engine = Pio.spf -> dimension -> size * partition * coarse_parameter
and grad_engine = Pio.spf -> fine_parameter -> fine_gradient
and optim_engine =
    Pio.spf -> partition -> coarse_parameter -> coarse_parameter * cost
and final_engine = Pio.spf -> fine_parameter -> unit
;;

type task_engines =
  {
   init_engine : init_engine;
   grad_engine : grad_engine;
   optim_engine : optim_engine;
   final_engine : final_engine;
  }
;;

type tasks =
  {
   init : (dimension, size * partition * coarse_parameter * cost) Pio.task;
   grad : (fine_parameter, fine_gradient) Pio.task;
   optim : (partition * coarse_parameter, coarse_parameter * cost) Pio.task;
   final : (fine_parameter, unit) Pio.task;
  }
;;

type check_valid_dimension =
  Upervasives.function_name -> dimension -> dimension -> unit
;;

type coarse_parameter_to_fine_parameter =
  partition -> coarse_parameter -> fine_parameter
;;

let coarse_parameter_to_fine_parameter ppm maa =
  Umatrix.map_ij (fun i _j pp_ij -> maa.(i).(Unatural.to_int pp_ij)) ppm
;;

let check_same_dimensions fun_name msg dim dim' =
  if dim' <> dim then w_invalid_arg fun_name msg
;;

let check_valid_dimension fun_name dim dim' =
  let msg = Printf.sprintf "dimension must be %d" (Unatural.to_int dim) in
  check_same_dimensions fun_name msg dim dim'
;;

let check_valid_zone_numbers fun_name ppm maa =
  let valid_zone_numbers =
    Uarray_ng.array_for_all_i
      (fun i ->
        Uarray.for_all
          (fun pp_ij -> pp_ij < Uarray.length (Uarray_ng.array_get maa i)))
      (Umatrix.to_array_array ppm) in
  if not valid_zone_numbers then
    w_invalid_arg fun_name "zone numbers in partition are not all valid"
;;

let make_init
    init_engine optim ({ Udebug.spf_debug = debug; _; } as spf) dim =
  debug (sprint_nat "dim" dim);
  let (sz, ppm1, maa1_init) = init_engine spf dim in
  debug (sprint_nat_couple "sz" sz);
  debug (sprint_nat_matrix "ppm1" ppm1);
  let (maa1, j1) = optim spf (ppm1, maa1_init) in
  debug (sprint_float_array_array "maa1" maa1);
  debug (sprint_float "j1" j1);
  (sz, ppm1, maa1, j1)
;;

let make_grad grad_engine ({ Udebug.spf_debug = debug; _; } as spf) pm =
  debug (sprint_float_matrix "pm" pm);
  let gm = grad_engine spf pm in
  debug (sprint_float_matrix "gm" gm);
  gm
;;

let make_optim
    optim_engine ({ Udebug.spf_debug = debug; _; } as spf) (ppm, maa_init) =
  debug (sprint_nat_matrix "ppm" ppm);
  debug (sprint_float_array_array "maa_init" maa_init);
  let dim = Uarray.length maa_init in
  let dim' = Umatrix.dim_i ppm in
  let msg =
    "arguments partition and initial coarse parameter \
     have incompatible dimensions" in
  check_same_dimensions "optim" msg dim dim';
  check_valid_zone_numbers "optim" ppm maa_init;
  let ((maa_opt, j_opt) as res) = optim_engine spf ppm maa_init in
  debug (sprint_float_array_array "maa_opt" maa_opt);
  debug (sprint_float "j_opt" j_opt);
  res
;;

let make_final
    final_engine ({ Udebug.spf_debug = debug; _; } as spf) pm =
  debug (sprint_float_matrix "pm" pm);
  final_engine spf pm;
;;

let make_tasks task_engines =
  let optim = make_optim task_engines.optim_engine in
  let init = make_init task_engines.init_engine optim in
  let grad = make_grad task_engines.grad_engine in
  let final = make_final task_engines.final_engine in
  { init = init;
    grad = grad;
    optim = optim;
    final = final; }
;;

(* Wrappers. *)

let mk_init_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    w_failwith "init_proj" "initialization should receive one value" else
  report "one value received";
  let dim = Nat.of_int (Typed_value.to_int vals.(0)) in
  report "dimension extracted";
  dim
;;

let mk_init_inj _spf ((ni, nj), ppm1, maa1, j1) =
  let sz = (Nat.to_int ni, Nat.to_int nj) in
  let ppm1 = Umatrix.map Nat.to_int ppm1 in
  let (lens1, a1) = Uarray_ng.array_array_pack maa1 in
  let lens1 = Array.map Nat.to_int lens1 in
  [| Typed_value.of_int_couple sz;
     Typed_value.of_c_int_matrix ppm1;
     Typed_value.of_int_vector lens1;
     Typed_value.of_float_vector a1;
     Typed_value.of_float j1; |]
;;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

let mk_grad_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    w_failwith "grad_proj" "gradient should receive one value" else
  report "one value received";
  let pm = Typed_value.to_c_float_matrix vals.(0) in
  report "fine parameter extracted";
  pm
;;

let mk_grad_inj _spf gm = [| Typed_value.of_c_float_matrix gm; |];;

let make_grad_wrapper =
  Pio.make_task_wrapper "grad" mk_grad_inj mk_grad_proj
;;

let mk_optim_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._3_ then
    w_failwith "optim_proj" "optimization should receive three values" else
  report "three values received";
  let ppm = Typed_value.to_c_int_matrix vals.(0) in
  let ppm = Umatrix.map Nat.of_int ppm in
  report "partition extracted";
  let lens_init = Typed_value.to_int_vector vals.(1) in
  let lens_init = Array.map Nat.of_int lens_init in
  report "coarse parameter lengths extracted";
  let a_init = Typed_value.to_float_vector vals.(2) in
  let maa_init = Uarray_ng.array_array_unpack lens_init a_init in
  report "initial coarse parameter extracted";
  (ppm, maa_init)
;;

let mk_optim_inj _spf (maa_opt, j_opt) =
  let (lens_opt, a_opt) = Uarray_ng.array_array_pack maa_opt in
  let lens_opt = Array.map Nat.to_int lens_opt in
  [| Typed_value.of_int_vector lens_opt;
     Typed_value.of_float_vector a_opt;
     Typed_value.of_float j_opt; |]
;;

let make_optim_wrapper =
  Pio.make_task_wrapper "optim" mk_optim_inj mk_optim_proj
;;

let mk_final_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    w_failwith "final_proj" "finalization should receive one value" else
  report "one value received";
  let pm = Typed_value.to_c_float_matrix vals.(0) in
  report "fine parameter extracted";
  pm
;;

let mk_final_inj _spf _ = [| Typed_value.of_string "OK, bye!"; |];;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;

(* Task computation functions. *)

let make_initialize init pf piob =
  Pio.worker_wrap_task piob (make_init_wrapper pf "init") init
;;

let make_compute_gradient grad pf piob =
  Pio.worker_wrap_task piob (make_grad_wrapper pf "grad") grad
;;

let make_optimize optim pf piob =
  Pio.worker_wrap_task piob (make_optim_wrapper pf "optim") optim
;;

let make_finalize final pf piob vals =
  Pio.worker_wrap_task piob (make_final_wrapper pf "final") final vals;
  (raise Exit : unit);
;;

(* Main loop. *)

let welcome_action () = "start user model"
and farewell_action () = "finish user model"
;;

let mk_task_tbl tasks =
  Ulist1.of_list
    [
     ("init", make_initialize tasks.init);
     ("grad", make_compute_gradient tasks.grad);
     ("optim", make_optimize tasks.optim);
     ("final", make_finalize tasks.final);
    ]
;;

let main_loop tasks =
  let task_tbl = mk_task_tbl tasks in
  Pio.worker_main welcome_action farewell_action task_tbl
;;
